/*
 *
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2021 Timm Fitschen <t.fitschen@indiscale.com>
 * Copyright (C) 2021 IndiScale GmbH <info@indiscale.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 */

#ifndef CONNECTION_H
#define CONNECTION_H
/**
 * @file connection.h
 * @author Timm Fitschen
 * @date 2021-05-18
 * @brief Configuration and setup of the connection.
 */
#include <string>
#include <iosfwd>
#include <memory>
#include <grpc/grpc.h>
#include <grpcpp/channel.h>
#include <grpcpp/client_context.h>
#include <grpcpp/create_channel.h>
#include <grpcpp/security/credentials.h>
#include "caosdb/info/v1alpha1/main.grpc.pb.h"

namespace caosdb {
using caosdb::info::v1alpha1::GeneralInfoService;
using caosdb::info::v1alpha1::VersionInfo;
using grpc::ChannelCredentials;

/**
 * @brief Configuration of the CaosDB connection.
 */
class CaosDBConnectionConfig {
  std::string host;
  int port;

public:
  CaosDBConnectionConfig(const std::string &host, int port);
  friend auto operator<<(std::ostream &out,
                         const CaosDBConnectionConfig &config)
    -> std::ostream &;

  [[nodiscard]] auto virtual toString() const -> std::string = 0;
  [[nodiscard]] auto getHost() const -> std::string;
  [[nodiscard]] auto getPort() const -> int;
  [[nodiscard]] auto virtual getChannelCredentials() const
    -> std::shared_ptr<ChannelCredentials> = 0;
};

class InsecureCaosDBConnectionConfig : public CaosDBConnectionConfig {
private:
  std::shared_ptr<ChannelCredentials> credentials;

public:
  InsecureCaosDBConnectionConfig(const std::string &host, int port);
  [[nodiscard]] auto toString() const -> std::string override;
  [[nodiscard]] auto getChannelCredentials() const
    -> std::shared_ptr<ChannelCredentials> override;
};

class SslCaosDBConnectionConfig : public CaosDBConnectionConfig {
private:
  std::shared_ptr<ChannelCredentials> credentials;

public:
  SslCaosDBConnectionConfig(const std::string &host, int port);
  SslCaosDBConnectionConfig(const std::string &host, int port,
                            const std::string &cacert);
};

/**
 * @brief A reusable connection to a CaosDBServer.
 */
class CaosDBConnection {
  std::shared_ptr<CaosDBConnectionConfig> config;
  std::unique_ptr<GeneralInfoService::Stub> stub_;

public:
  explicit CaosDBConnection(
    const std::shared_ptr<CaosDBConnectionConfig> &config);
  friend auto operator<<(std::ostream &out, const CaosDBConnection &connection)
    -> std::ostream &;
  [[nodiscard]] auto getGeneralInfoService() const
    -> GeneralInfoService::Stub &;
  [[nodiscard]] auto getVersionInfo() const -> const VersionInfo &;
};
} // namespace caosdb
#endif
